@@ -12,11 +12,12 @@ from dataclasses import dataclass
 from typing import List, Optional, Union
 from azure.ml.component import dsl
 from azure.ml.component.pipeline import Pipeline
-from azure.ml.component.component import Component
+from azure.ml.component.component import Component, Output
 import logging
 from omegaconf import OmegaConf, open_dict, DictConfig
 from pathlib import Path
-from typing import Callable
+from typing import Callable, Dict
+import collections.abc
 
 log = logging.getLogger(__name__)
 
@@ -27,12 +28,12 @@ EXPERIMENTAL_WARNING_MSG = (
 
 def dataset_name_to_reference(
     component_instance: Component, dataset_names: list
-) -> list:
-    res = []
+) -> Dict[str, Output]:
+    res = {}
     if not dataset_names:
         return res
     for dataset in dataset_names:
-        res.append(component_instance.outputs[dataset])
+        res[dataset] = component_instance.outputs[dataset]
     return res
 
 
@@ -65,6 +66,7 @@ class FederatedPipelineBase(AMLPipelineHelper):
             self.module_loader.modules_manifest, {"fl_fedavg_pytorch": fedavg_component}
         )
         self.cur_fl_iteration: Optional[int] = None
+        self.output_prefix = "fl_output_"
 
     @experimental(message=EXPERIMENTAL_WARNING_MSG)
     def create_base_name(self) -> str:
@@ -78,55 +80,47 @@ class FederatedPipelineBase(AMLPipelineHelper):
         rv += generate_slug(2)
         return rv
 
-    def _expand_input_list_to_component_args_and_apply_to_fedavg(
-        self,
-        component: Callable,
-        num_models: int,
-        input: list,
-        weights: str,
-        model_name: str,
-    ) -> Component:
-        input_mapping = {}
-        for i in range(num_models):
-            input_name = "model_input_dir_" + str(i + 1)
-            try:
-                input_mapping[input_name] = input[i]
-            except IndexError as e:
-                log.error(
-                    "The input model weights are inconsistent with the number of models."
-                )
-                raise e
-        step = component(
-            **input_mapping,
-            num_models=num_models,
-            weights=weights,
-            model_name=model_name,
-        )
-        return step
+    @experimental(message=EXPERIMENTAL_WARNING_MSG)
+    def create_iteration_name(self) -> str:
+        """Display name for each iteration's subgraph, containing letters, numbers, and underscores only. The default value is "iteration n"."""
+        return f"iteration_{self.cur_fl_iteration}"
 
     @experimental(message=EXPERIMENTAL_WARNING_MSG)
-    def preprocess(
-        self, config: DictConfig, input: Optional[list] = None
-    ) -> StepOutput:
+    def create_silo_specific_name(self, silo_name) -> str:
+        """Display name for a silo's training in iteration n's subgraph, containing letters, numbers, and underscores only. The default value is "silo_<silo_name>__iteration_n"."""
+        return f"silo_{silo_name}__iteration_{self.cur_fl_iteration}"
+
+    @experimental(message=EXPERIMENTAL_WARNING_MSG)
+    def create_subgraph_name(self) -> str:
+        """Display name for each subgraph"""
+        return "Federated Learning Subgraph"
+
+    @experimental(message=EXPERIMENTAL_WARNING_MSG)
+    def preprocess(self, config: DictConfig, **input) -> StepOutput:
         """
         Optional user-defined preprocess step. The outputs will be distributed to each silo's datastore.
 
         Returns:
-            a component/subgraph instance, and a list of output dataset name to be passed to the downstream pipeline **in order**.
+            a component/subgraph instance, and a list of output dataset names to be passed to the downstream pipeline.
         """
         pass
 
     @experimental(message=EXPERIMENTAL_WARNING_MSG)
-    def midprocess(self, config: DictConfig, input: list) -> StepOutput:
+    def midprocess(self, config: DictConfig, **input) -> StepOutput:
         """
         User-defined midprocess step which reads outputs from `train` in each silo. The outputs will be distributed to each silo's datastore.
         By default, it calls the built-in FedAvg component (un-signed).
 
         Returns:
-            a component/subgraph instance, and a list of output dataset name to be passed to the downstream pipeline **in order**.
+            a component/subgraph instance, and a list of output dataset name to be passed to the downstream pipeline.
         """
         fedavg_component = self.component_load("fl_fedavg_pytorch")
         num_models = len(self.config.federated_config.silos)
+        if len(input) != num_models:
+            log.error(
+                "The input model weights are inconsistent with the number of models."
+            )
+            raise Exception
         if "agg_weights" in self.config.federated_config.params:
             weights = self.config.federated_config.params.agg_weights
         else:
@@ -136,34 +130,43 @@ class FederatedPipelineBase(AMLPipelineHelper):
             if "model_name" in self.config.federated_config.params
             else None
         )
-        fedavg_step = self._expand_input_list_to_component_args_and_apply_to_fedavg(
-            fedavg_component, num_models, input, weights, model_name
+        updated_input = {}
+        for k, v in input.items():
+            idx = k[k.index("silo") + 4 : k.index("_")]
+            updated_input["model_input_dir_" + idx] = v
+        fedavg_step = fedavg_component(
+            **updated_input,
+            num_models=num_models,
+            weights=weights,
+            model_name=model_name,
         )
         return StepOutput(fedavg_step, ["model_output_dir"])
 
     @experimental(message=EXPERIMENTAL_WARNING_MSG)
-    def postprocess(self, config: DictConfig, input: list) -> StepOutput:
+    def postprocess(self, config: DictConfig, **kwargs) -> StepOutput:
         """
-        Optional user-defined postprocess step which reads outputs from `train` in each silo and writes to the `noncompliant_datastore`.
+        Optional user-defined postprocess step which reads outputs from the final `midprocess` step and writes to the `noncompliant_datastore`.
 
         Returns:
-            a component/subgraph instance, and a list of output dataset name to be passed to the downstream pipeline **in order**.
+            a component/subgraph instance, and a list of output dataset names.
         """
         pass
 
     @experimental(message=EXPERIMENTAL_WARNING_MSG)
-    def train(self, config: DictConfig, input: list, silo: DictConfig) -> StepOutput:
+    def train(
+        self, config: DictConfig, silo: DictConfig, input: Output = None, **kwargs
+    ) -> StepOutput:
         """
         User-defined train step happening at each silo. This reads outputs from `preprocess` or `midprocess`, and sends outputs back to the `noncompliant_datastore`.
 
         Returns:
-            a component/subgraph instance, and a list of output dataset name to be passed to the downstream pipeline **in order**.
+            a component/subgraph instance, and a list of output dataset names to be passed to the downstream pipeline.
         """
         pass
 
     def _data_transfer(
-        self, input_list: list, destination_datastore: str, base_name: str
-    ) -> list:
+        self, input_list: dict, destination_datastore: str, base_name: str
+    ) -> dict:
         """Moves data `input_list` to `destination_datastore` with path prefixed by `base_name`."""
         # assume one shared ADF
         dts_name = self.config.federated_config.data_transfer_component
@@ -178,26 +181,45 @@ class FederatedPipelineBase(AMLPipelineHelper):
             )
             raise (e)
 
-        res = []
+        res = {}
         if not input_list:
             return res
-        for input in input_list:
+        for name, input in input_list.items():
             data_transfer_step = data_transfer_component(source_data=input)
             self.apply_smart_runsettings(data_transfer_step, datatransfer=True)
             data_transfer_step.outputs.destination_data.configure(
                 datastore=destination_datastore,
-                path_on_datastore=base_name + "/" + input.port_name,
+                path_on_datastore=base_name + "/" + name,
             )
-            res.append(data_transfer_step.outputs.destination_data)
+            res[name] = data_transfer_step.outputs.destination_data
         return res
 
-    def _process_at_orchestrator(self, prev: StepOutput) -> list:
+    def _update_nested_dict_with_non_none_values(self, dict1, dict2):
+        """Update dict1 with non-none values in dict2"""
+        for k, v in dict2.items():
+            if not v:
+                continue
+            if isinstance(v, collections.abc.Mapping):
+                dict1[k] = self._update_nested_dict_with_non_none_values(
+                    dict1.get(k, {}), v
+                )
+            else:
+                dict1[k] = v
+        return dict1
+
+    def _process_at_orchestrator(self, prev: StepOutput) -> dict:
         """Processes outputs at the orchestrator."""
-        if isinstance(prev.step, Pipeline):
-            prev_output = list(prev.step.outputs.values())
-        elif isinstance(prev.step, Component):
-            self.apply_smart_runsettings(prev.step)
-            prev_output = dataset_name_to_reference(prev.step, prev.outputs)
+        step = prev.step
+        if isinstance(step, Pipeline):
+            prev_output = step.outputs
+        elif isinstance(step, Component):
+            runsettings = step.runsettings._get_values()
+            self.apply_smart_runsettings(step)
+            merged_runsettings = self._update_nested_dict_with_non_none_values(
+                step.runsettings._get_values(), runsettings
+            )
+            step.runsettings._set_section_by_dict(merged_runsettings)
+            prev_output = dataset_name_to_reference(step, prev.outputs)
         else:
             raise Exception(
                 f"The output of preprocess/midprocess/postprocess step must be a Pipeline object or a Component object. You are using {type(prev.step)}."
@@ -233,48 +255,59 @@ class FederatedPipelineBase(AMLPipelineHelper):
             )
 
     def _check_if_data_transfer_is_activated_and_process(
-        self, prev_output, datastore_name, base_name
+        self, datastore_name, base_name, **inputs
     ):
         deactivate_data_transfer = self.config.federated_config.deactivate_data_transfer
         if deactivate_data_transfer:
             log.info("Data transfer is disabled; training outputs will remain in silo.")
-            return prev_output
+            return inputs
         else:
-            log.info(f"Moving data {prev_output} into {datastore_name}.")
+            log.info(f"Moving data {inputs.keys()} into {datastore_name}.")
             return self._data_transfer(
-                input_list=prev_output,
+                input_list=inputs,
                 destination_datastore=datastore_name,
                 base_name=base_name,
             )
 
     def _train_in_silo_once(
-        self, silo_name: str, silo: DictConfig, prev_output: list, cool_name: str
-    ) -> list:
+        self, silo_name: str, silo: DictConfig, cool_name: str, **prev_output
+    ) -> dict:
         """Runs the "training" step once at one silo."""
         prev_output = self._check_if_data_transfer_is_activated_and_process(
-            prev_output, silo.datastore, cool_name
+            silo.datastore, cool_name, **prev_output
         )
-        train_step = self.train(self.config, input=prev_output, silo=silo)
-        if isinstance(train_step.step, Pipeline):
-            train_output = list(train_step.step.outputs.values())
-        elif isinstance(train_step.step, Component):
+        if len(prev_output) == 1:
+            prev_output = {"input": list(prev_output.values())[0]}
+        train_step = self.train(self.config, silo=silo, **prev_output)
+
+        step = train_step.step
+        if isinstance(step, Pipeline):
+            train_output = step.outputs
+            for node in step._expand_pipeline_nodes():
+                node._node_name = self.create_silo_specific_name(silo_name)
+        elif isinstance(step, Component):
+            runsettings = step.runsettings._get_values()
             self.apply_smart_runsettings(
-                train_step.step,
+                step,
                 target=silo.compute,
                 datastore_name=silo.datastore,
             )
-            train_output = dataset_name_to_reference(
-                train_step.step, train_step.outputs
+            merged_runsettings = self._update_nested_dict_with_non_none_values(
+                step.runsettings._get_values(), runsettings
             )
+            step.runsettings._set_section_by_dict(merged_runsettings)
+            train_output = dataset_name_to_reference(step, train_step.outputs)
+            step._node_name = self.create_silo_specific_name(silo_name)
+
         else:
             raise Exception(
-                f"The output of train step must be a Pipeline object or a Component object. You are using {type(train_step.step)}"
+                f"The output of train step must be a Pipeline object or a Component object. You are using {type(step)}"
             )
 
         silo_output = self._check_if_data_transfer_is_activated_and_process(
-            train_output,
             self.config.compute.noncompliant_datastore,
             cool_name + "/" + silo_name,
+            **train_output,
         )
         return silo_output
 
@@ -282,6 +315,33 @@ class FederatedPipelineBase(AMLPipelineHelper):
     def build(self, config: DictConfig):
         """Constructs the federated pipeline. User does not need to modify."""
 
+        @dsl.pipeline(display_name=self.create_subgraph_name())
+        def _subpipeline_function(**kwargs):
+            cool_name = self.create_base_name()
+            midprocess_input = {}
+            i = 1
+            for silo_name, silo in self.config.federated_config.silos.items():
+                train_output = self._train_in_silo_once(
+                    silo_name,
+                    silo,
+                    cool_name,
+                    **kwargs,
+                )
+                if len(train_output) == 1:
+                    # if there's only one output, forcibly change the name to "<silo_name>_input"
+                    midprocess_input[f"silo{i}_input"] = list(train_output.values())[0]
+                else:
+                    for k, v in train_output.items():
+                        midprocess_input[f"silo{i}_" + k] = v
+                i += 1
+
+            prev = self.midprocess(self.config, **midprocess_input)
+            prev_output = self._process_at_orchestrator(prev) if prev else None
+            output = {
+                self.output_prefix + name: prev_output[name] for name in prev_output
+            }
+            return output
+
         @dsl.pipeline()
         def pipeline_function():
             self._merge_config()
@@ -290,19 +350,24 @@ class FederatedPipelineBase(AMLPipelineHelper):
 
             for iter in range(self.config.federated_config.max_iterations):
                 self.cur_fl_iteration = iter
-                cool_name = self.create_base_name()
-                midprocess_input = []
-                for silo_name, silo in self.config.federated_config.silos.items():
-                    train_output = self._train_in_silo_once(
-                        silo_name, silo, prev_output, cool_name
-                    )
-                    midprocess_input += train_output
+                if prev_output:
+                    prev_output = {
+                        output_name.replace(self.output_prefix, ""): prev_output[
+                            output_name
+                        ]
+                        for output_name in prev_output.keys()
+                    }
+                    subpipeline = _subpipeline_function(**prev_output)
+                else:
+                    subpipeline = _subpipeline_function()
+                subpipeline.node_name = self.create_iteration_name()
+                prev_output = subpipeline.outputs
 
-                prev = self.midprocess(self.config, input=midprocess_input)
-                prev_output = self._process_at_orchestrator(prev) if prev else None
-
-            prev = self.postprocess(self.config, input=prev_output)
-            prev_output = self._process_at_orchestrator(prev) if prev else None
+            if len(prev_output) == 1:
+                prev_output = {"input": list(prev_output.values())[0]}
+            prev = self.postprocess(self.config, **prev_output)
+            output = self._process_at_orchestrator(prev) if prev else None
+            return output
 
         return pipeline_function
 
