@@ -66,6 +66,7 @@ class pipeline_compute_config:  # pylint: disable=invalid-name
     windows_output_mode: str = "upload"
 
     hdi_prod_target: str = MISSING
+    hdi_dc_target: Optional[str] = None
     hdi_driver_memory: str = "4g"
     hdi_driver_cores: int = 2
     hdi_executor_memory: str = "3g"
@@ -84,6 +85,7 @@ class pipeline_compute_config:  # pylint: disable=invalid-name
 
     compliant_datastore: str = MISSING
     noncompliant_datastore: Optional[str] = MISSING
+    dc_datastore: Optional[str] = None
 
 
 @dataclass
