@@ -5,6 +5,7 @@
 Pipeline helper class to create pipelines loading modules from a flexible manifest.
 """
 import argparse
+from collections import defaultdict
 import os
 import json
 import logging
@@ -17,8 +18,10 @@ import jsonpath_ng
 from typing import Callable
 from hashlib import sha256
 from pathlib import Path
+from toposort import toposort_flatten
 import datetime
 
+
 try:
     import hydra
     from hydra.core.config_store import ConfigStore
@@ -36,6 +39,7 @@ try:
         ComponentDefinition,
         ComponentType,
     )
+    from azure.ml.component.pipeline import PipelineParameter
 except ImportError as error:
     raise ImportError(
         f"{error.msg}. Please install using `pip install shrike[pipeline]`."
@@ -388,8 +392,14 @@ class AMLPipelineHelper:
             output_mode (str): force output_mode over hydra conf
             custom_runtime_arguments (dict): any additional custom args
         """
+        if self.module_loader.is_local(module_name):
+            target = target if target is not None else self.config.compute.hdi_dc_target
+        else:
+            target = (
+                target if target is not None else self.config.compute.hdi_prod_target
+            )
         log.info(
-            f"Using HDI compute target {self.config.compute.hdi_prod_target} to run {module_name} from pipeline class {self.__class__.__name__}"
+            f"Using HDI compute target {target} to run {module_name} from pipeline class {self.__class__.__name__}"
         )
         if custom_runtime_arguments:
             log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")
@@ -1073,16 +1083,19 @@ class AMLPipelineHelper:
             return
 
         if windows:
-            # no detonation chamber, we an't use "use_local" here
             self._apply_windows_runsettings(
-                module_name, module_instance, mpi=mpi, **custom_runtime_arguments
+                module_name,
+                module_instance,
+                mpi=mpi,
+                **custom_runtime_arguments,
             )
             return
 
         if hdi:
-            # no detonation chamber, we an't use "use_local" here
             self._apply_hdi_runsettings(
-                module_name, module_instance, **custom_runtime_arguments
+                module_name,
+                module_instance,
+                **custom_runtime_arguments,
             )
             return
 
@@ -1099,7 +1112,11 @@ class AMLPipelineHelper:
             return
 
         self._apply_linux_runsettings(
-            module_name, module_instance, mpi=mpi, gpu=gpu, **custom_runtime_arguments
+            module_name,
+            module_instance,
+            mpi=mpi,
+            gpu=gpu,
+            **custom_runtime_arguments,
         )
 
     def _parse_pipeline_tags(self):
@@ -1392,6 +1409,71 @@ class AMLPipelineHelper:
             except BaseException as e:
                 log.error(f"An error occurred, recovery is not successful: {e}")
 
+    def _update_hdi_dc_configs(self, pipeline):
+        nodes = pipeline._expand_pipeline_nodes()
+        graph = defaultdict(list)
+        node_mapping = {}
+        for n in nodes:
+            node_mapping[n._instance_id] = n
+            for input in n.inputs.values():
+                try:
+                    owner = input._get_data_owner()
+                    if owner:
+                        graph[n._instance_id].append(owner._instance_id)
+                except AttributeError:
+                    continue
+        sorted_node_ids = toposort_flatten(graph)
+        use_hdi_dc = defaultdict(bool)
+        for id in sorted_node_ids:
+            if id not in node_mapping:
+                continue
+            node = node_mapping[id]
+            if node.runsettings.target == self.config.compute.hdi_dc_target:
+                use_hdi_dc[id] = True
+                continue
+            for parent in graph[id]:
+                if use_hdi_dc[parent]:
+                    use_hdi_dc[id] = True
+                    break
+
+        for node in nodes:
+            if use_hdi_dc[node._instance_id]:
+                self._set_all_outputs_to(
+                    node,
+                    output_mode=None,
+                    datastore_name=self.config.compute.dc_datastore,
+                )
+                log.info(
+                    f"Writing outputs of {node.name} to dc_datastore {self.config.compute.dc_datastore}"
+                )
+                if node.runsettings.target == self.config.compute.hdi_prod_target:
+                    node.runsettings.configure(target=self.config.compute.hdi_dc_target)
+                elif (
+                    node.runsettings.target == self.config.compute.linux_cpu_prod_target
+                ):
+                    node.runsettings.configure(
+                        target=self.config.compute.linux_cpu_dc_target
+                    )
+                elif (
+                    node.runsettings.target == self.config.compute.linux_gpu_prod_target
+                ):
+                    node.runsettings.configure(
+                        target=self.config.compute.linux_gpu_dc_target
+                    )
+                elif (
+                    node.runsettings.target
+                    == self.config.compute.windows_cpu_prod_target
+                ):
+                    node.runsettings.configure(
+                        target=self.config.compute.windows_cpu_dc_target
+                    )
+                else:
+                    continue
+                log.info(
+                    f"Updating compute target for {node.name} {node._instance_id} to {node.runsettings.target}"
+                )
+        return pipeline
+
     def _publish_to_endpoint(self, pipeline):
         name = datetime.datetime.now()
         endpoint_name = self.config.run.endpoint_name or self.config.run.experiment_name
@@ -1437,6 +1519,9 @@ class AMLPipelineHelper:
         log.info("Creating Pipeline Instance...")
         pipeline = self.pipeline_instance(pipeline_function, self.config)
 
+        log.info("Overriding compute target if upstream steps use HDI DC...")
+        pipeline = self._update_hdi_dc_configs(pipeline)
+
         log.info("Validating...")
         pipeline.validate(workspace=current_workspace())
 
